/*
** ucview.c
** 
** Made by (Arne Caspari)
** Login   <arne@arne-laptop>
** 
** Started on  Mon Nov  6 07:22:56 2006 Arne Caspari
** Last update Fri Mar 16 18:38:37 2007 Arne Caspari
*/

/*
  Copyright (C) 2007  Arne Caspari <arne@unicap-imaging.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  Main application window and mainloop / callback handling

  TODO: 

   - i18n
   - load icons from files 
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <unicap.h>
#include <unicapgtk.h>
#include <ucil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ucview.h"
#include "device_dialog.h"
#include "settings_dialog.h"
#include "checkperm.h"
#include "image_effects.h"
#include "callbacks.h"
#include "gui.h"

#include "icon-fps.h"
#include "icon-measure.h"
#include "icon-colormag.h"
#include "icon-record.h"
#include "icon-play.h"
#include "icon-still-image.h"
#include "icon-devsettings.h"
#include "icon-ucview.h"


static void ucview_shutdown( unicap_handle_t handle )
{
   gdk_threads_leave();
   if( handle )
   {
      unicap_stop_capture( handle );
      unicap_close( handle );
   }
   gdk_threads_enter();

   gtk_main_quit();
}

static void predisplay_cb( UnicapgtkVideoDisplay *ugtk, unicap_data_buffer_t *buffer, new_frame_data_t *data )
{
   GList *entry;
   
   if( data->record )
   {
      ucil_encode_frame( data->video_object, buffer );
   }

   if( data->still_image )
   {
      GdkPixbuf *pixbuf;

      ucil_convert_buffer( &data->still_image_buffer, buffer );
      
      
      pixbuf = gdk_pixbuf_new_from_data( data->still_image_buffer.data, 
					 GDK_COLORSPACE_RGB, 
					 0, 
					 8, 
					 data->still_image_buffer.format.size.width, 
					 data->still_image_buffer.format.size.height, 
					 data->still_image_buffer.format.size.width * 3, 
					 NULL, 
					 NULL );
					 
					 
      if( pixbuf )
      {
	 gdk_pixbuf_save( pixbuf, data->still_image_path, "jpeg", NULL, NULL );
	 g_object_unref( pixbuf );
      }
      
      data->still_image = FALSE;
   }

   g_mutex_lock( data->mutex );
   for( entry = g_list_first( data->image_effects );
	entry;
	entry = g_list_next( entry ) )
   {
      image_effect_data_t *effect_data = ( image_effect_data_t * ) entry->data;
      effect_data->func( buffer, effect_data->data );
   }
   g_mutex_unlock( data->mutex );
}

GtkWidget *create_display_widget( unicap_handle_t handle )
{
   GtkWidget *ugtk;
   unicap_format_t format;

   unicap_get_format( handle, &format );
   format.buffer_type = UNICAP_BUFFER_TYPE_SYSTEM;
   unicap_set_format( handle, &format );

   ugtk = unicapgtk_video_display_new( );  

   return ugtk;
}


GtkWidget *create_main_window( new_frame_data_t *new_frame_data, unicap_handle_t handle )
{
   GtkWidget *window;
   GtkWidget *vbox;
   GtkWidget *toolbar;
   GtkWidget *scrolled_window;
   GtkToolItem *button;
   GdkPixbuf *pixbuf;
   GtkWidget *property_dialog;
   GtkWidget *settings_dialog;
   GtkUIManager *ui;

   vbox = gtk_vbox_new( FALSE, 0 );
   
   window = gtk_window_new( GTK_WINDOW_TOPLEVEL );   
   gtk_container_add( GTK_CONTAINER( window ), vbox );
   pixbuf = gdk_pixbuf_new_from_inline( -1, icon_ucview_data, FALSE, NULL );
   gtk_window_set_icon( GTK_WINDOW( window ), pixbuf );
   g_object_unref( G_OBJECT( pixbuf ) );
   
   ui = gui_create_manager( new_frame_data );
   gtk_box_pack_start( GTK_BOX( vbox ), gtk_ui_manager_get_widget( ui, "/MenuBar" ), FALSE, FALSE, 0 );
   gtk_window_add_accel_group( GTK_WINDOW( window ), gtk_ui_manager_get_accel_group( ui ) );

/*    toolbar = gtk_toolbar_new(); */
   toolbar = gtk_ui_manager_get_widget( ui, "/ToolBar" );
   gtk_box_pack_start( GTK_BOX( vbox ), toolbar, FALSE, FALSE, 0 );

/*    // FPS */
/*    button = gtk_toggle_tool_button_new(); */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_fps_data, FALSE, NULL ); */
/*    gtk_tool_button_set_label( GTK_TOOL_BUTTON( button ), "FPS" ); */
/*    gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON( button ), gtk_image_new_from_pixbuf( pixbuf ) ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "toggled", G_CALLBACK( fps_toggled_cb ), new_frame_data ); */

/*    // Measure */
/*    button = gtk_toggle_tool_button_new(); */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_measure_data, FALSE, NULL ); */
/*    gtk_tool_button_set_label( GTK_TOOL_BUTTON( button ), "Measure" ); */
/*    gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON( button ), gtk_image_new_from_pixbuf( pixbuf ) ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "toggled", G_CALLBACK( measure_toggled_cb ), new_frame_data ); */

/*    // Color Magnifier */
/*    button = gtk_toggle_tool_button_new(); */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_colormag_data, FALSE, NULL ); */
/*    gtk_tool_button_set_label( GTK_TOOL_BUTTON( button ), "Color Magnifier" ); */
/*    gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON( button ), gtk_image_new_from_pixbuf( pixbuf ) ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "toggled", G_CALLBACK( color_mag_toggled_cb ), new_frame_data ); */

/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), gtk_separator_tool_item_new( ), -1 );    */

/*    // Pause */
/*    button = gtk_toggle_tool_button_new_from_stock( GTK_STOCK_MEDIA_PAUSE ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_object_set_data( G_OBJECT( window ), "pause button", button ); */

/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), gtk_separator_tool_item_new( ), -1 ); */

/*    // Record */
/*    button = gtk_toggle_tool_button_new(); */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_record_data, FALSE, NULL ); */
/*    gtk_tool_button_set_label( GTK_TOOL_BUTTON( button ), "Record Video" ); */
/*    gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON( button ), gtk_image_new_from_pixbuf( pixbuf ) ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "toggled", G_CALLBACK( record_toggled_cb ), new_frame_data ); */
   
/*    button = gtk_toggle_tool_button_new(); */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_play_data, FALSE, NULL ); */
/*    gtk_tool_button_set_label( GTK_TOOL_BUTTON( button ), "Play" ); */
/*    gtk_tool_button_set_icon_widget( GTK_TOOL_BUTTON( button ), gtk_image_new_from_pixbuf( pixbuf ) ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "toggled", G_CALLBACK( play_toggled_cb ), new_frame_data ); */
   

/*    // Still Image */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_still_image_data, FALSE, NULL ); */
/*    button = gtk_tool_button_new(gtk_image_new_from_pixbuf( pixbuf ), "Save Still Image" ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 ); */
/*    g_signal_connect( G_OBJECT( button ), "clicked", G_CALLBACK( save_still_image_cb ), new_frame_data ); */

/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), gtk_separator_tool_item_new( ), -1 ); */

/*    // Device Settings */
/*    pixbuf = gdk_pixbuf_new_from_inline( -1, icon_devsettings_data, FALSE, NULL ); */
/*    button = gtk_tool_button_new( gtk_image_new_from_pixbuf( pixbuf ), "Device Settigns" ); */
/*    g_object_unref( G_OBJECT( pixbuf ) ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 );    */
   property_dialog = unicapgtk_property_dialog_new_by_handle( handle );
   g_signal_connect_swapped( G_OBJECT( window ), "destroy", G_CALLBACK( gtk_widget_destroy ), property_dialog );
/*    g_signal_connect_swapped( G_OBJECT( button ), "clicked", G_CALLBACK( gtk_window_present ), property_dialog ); */
   gtk_window_set_transient_for( GTK_WINDOW( property_dialog ), GTK_WINDOW( window ) );
   new_frame_data->property_dialog = property_dialog;

/*    // Application Settings */
/*    button = gtk_tool_button_new_from_stock( GTK_STOCK_PREFERENCES ); */
/*    gtk_toolbar_insert( GTK_TOOLBAR( toolbar ), GTK_TOOL_ITEM( button ), -1 );    */

   settings_dialog = settings_dialog_new();
   g_signal_connect_swapped( G_OBJECT( window ), "destroy", G_CALLBACK( gtk_widget_destroy ), settings_dialog );
/*    g_signal_connect_swapped( G_OBJECT( button ), "clicked", G_CALLBACK( gtk_window_present ), settings_dialog ); */
   g_signal_connect( G_OBJECT( settings_dialog ), "delete-event", G_CALLBACK( gtk_widget_hide_on_delete ), NULL );   
   gtk_window_set_transient_for( GTK_WINDOW( settings_dialog ), GTK_WINDOW( window ) );
   new_frame_data->settings_dialog = settings_dialog;
   
   scrolled_window = gtk_scrolled_window_new( NULL, NULL );
   gtk_box_pack_start( GTK_BOX( vbox ), scrolled_window, TRUE, TRUE, 0 );
   gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( scrolled_window ), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
   g_object_set_data( G_OBJECT( window ), "display_window", scrolled_window );
   
   return window;
}


int main( int argc, char**argv )
{
   GtkWidget *window;
   GtkWidget *display;
   GtkWidget *display_window;
   GtkWidget *pause_button;
   GConfClient *client;


   device_dialog_t *device_dialog;
   unicap_handle_t handle;
   unicap_format_t format;
   unicap_device_t device;
   
   new_frame_data_t new_frame_data;

   g_thread_init(NULL);
   gdk_threads_init();
   gdk_threads_enter();
   gtk_init (&argc, &argv);

   client = gconf_client_get_default();
   gconf_client_add_dir( client, UCVIEW_GCONF_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL );
   check_unicap_env( client );
   
   new_frame_data.client = client;
   new_frame_data.image_effects = NULL;
   new_frame_data.mutex = g_mutex_new();
   new_frame_data.record = FALSE;
   new_frame_data.still_image = FALSE;
   new_frame_data.last_video_file = NULL;

   if( !SUCCESS( unicap_enumerate_devices( NULL, &device, 0 ) ) )
   {
      GtkWidget *dlg;
      dlg = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
				    "No supported video capture device found!" );
      gtk_dialog_run( GTK_DIALOG( dlg ) );
      return -1;
   }
/*    else if( !SUCCESS( unicap_enumerate_devices( NULL, &device, 1 ) ) ) */
/*    { */
/*       unicap_enumerate_devices( NULL, &device, 0 ); */
/*       if( !SUCCESS( unicap_open( &handle, &device ) ) ) */
/*       { */
/* 	 GtkWidget *dlg; */
/* 	 dlg = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,  */
/* 				       "Failed to open device '%s'\nand only one device found!", device.identifier ); */
/* 	 gtk_dialog_run( GTK_DIALOG( dlg ) ); */
/* 	 return -1; */
/*       } */
      
/*    } */
   else
   {
      device_dialog = create_device_dialog(FALSE, NULL);
      handle = device_dialog_run( device_dialog );
      device_dialog_destroy( device_dialog );
   }
   if( !handle )
   {
      GtkWidget *dlg;
      dlg = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
				    "No device!" );
      gtk_dialog_run( GTK_DIALOG( dlg ) );
      return -1;
   }

   ifx_init();

   new_frame_data.device_handle = handle;

   window = create_main_window( &new_frame_data, handle );
   
   display = create_display_widget( handle );
   display_window = g_object_get_data( G_OBJECT( window ), "display_window" );
   new_frame_data.display = display;
   g_assert( display_window );
/*    pause_button = g_object_get_data( G_OBJECT( window ), "pause button" ); */
/*    g_assert( pause_button ); */
/*    g_signal_connect( G_OBJECT( pause_button ), "toggled", G_CALLBACK( pause_state_toggled_cb), display ); */

   unicap_get_format( handle, &format );
   if( ( format.size.width > 1024 ) || ( format.size.height > 768 ) )
   {
      gtk_widget_set_size_request( display_window, 1024, 768 );
   }
   else
   {
      gtk_widget_set_size_request( display_window, format.size.width + 10, format.size.height + 10 );
   }
   gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW( display_window ), display );
   gtk_widget_show_all( window );
   g_signal_connect_swapped( G_OBJECT (window), "delete-event", (GCallback)ucview_shutdown, handle );
   unicapgtk_video_display_set_handle( UNICAPGTK_VIDEO_DISPLAY( display ), handle );
   printf( "format size: %dx%d\n", format.size.width, format.size.height );
   unicapgtk_video_display_set_format( UNICAPGTK_VIDEO_DISPLAY( display ), &format );
   unicapgtk_video_display_start( UNICAPGTK_VIDEO_DISPLAY( display ) );

   unicap_copy_format( &new_frame_data.format, &format );
   unicap_copy_format( &new_frame_data.still_image_buffer.format, &format );
   new_frame_data.still_image_buffer.format.fourcc = UCIL_FOURCC( 'R', 'G', 'B', '3' );
   new_frame_data.still_image_buffer.format.bpp = 24;
   new_frame_data.still_image_buffer.format.buffer_size = new_frame_data.still_image_buffer.format.size.width * 
      new_frame_data.still_image_buffer.format.size.height * 3;
   new_frame_data.still_image_buffer.buffer_size = new_frame_data.still_image_buffer.format.buffer_size;
   new_frame_data.still_image_buffer.data = malloc( new_frame_data.still_image_buffer.format.buffer_size );

   g_signal_connect( G_OBJECT( display ), "motion_notify_event", G_CALLBACK( display_motion_notify_cb ), &new_frame_data );
   g_signal_connect( G_OBJECT( display ), "button_press_event", G_CALLBACK( display_button_press_event_cb ), &new_frame_data );
   g_signal_connect( G_OBJECT( display ), "unicapgtk_video_display_predisplay", G_CALLBACK( predisplay_cb ), &new_frame_data );
   
   gtk_main();

   g_object_unref( G_OBJECT( client ) );
   gdk_threads_leave();
   return 0;
}
